'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _levelup = require('levelup');

var _levelup2 = _interopRequireDefault(_levelup);

var _mongodown = require('mongodown');

var _mongodown2 = _interopRequireDefault(_mongodown);

var _mongojs = require('mongojs');

var _mongojs2 = _interopRequireDefault(_mongojs);

var _levelgraph = require('levelgraph');

var _levelgraph2 = _interopRequireDefault(_levelgraph);

var _levelgraphRecursive = require('levelgraph-recursive');

var _levelgraphRecursive2 = _interopRequireDefault(_levelgraphRecursive);

var _expand = require('./expand');

var _expand2 = _interopRequireDefault(_expand);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const clean = function clean(dbName, cb) {
  const db = (0, _mongojs2.default)(dbName);
  db.dropDatabase(err => {
    db.close();
    cb(err);
  });
};

const create = function create(dbName, cleanDb, cb) {
  const postClean = () => (0, _levelup2.default)(dbName, { db: _mongodown2.default }, (err, leveldb) => {
    if (err) {
      return cb(err);
    }
    const db = (0, _levelgraphRecursive2.default)((0, _levelgraph2.default)(leveldb));
    return cb(err, (0, _expand2.default)(db, leveldb));
  });

  if (cleanDb) {
    return clean(dbName, err => {
      if (err) {
        return cb(err);
      }
      return postClean();
    });
  }
  return postClean();
};

// Same as create except you can load additional data into the system
const load = function load(dbName, files, cleanDb, cb) {
  create(dbName, cleanDb, (err, db) => {
    if (err) {
      return cb(err);
    }
    return db.loadFiles(files, err2 => {
      cb(err2, db);
    });
  });
};

exports.default = {
  clean,
  create,
  load
};